'''
Send text to OBS Studio
'''

import asyncio
import logging
import simpleobsws
from slidetextbridge.core import config
from . import base

class ObsWsEmitter(base.PluginBase):
    '''
    Emit text to OBS Studio
    '''
    @staticmethod
    def type_name():
        return 'obsws'

    @staticmethod
    def config(data):
        'Return the config object'
        cfg = config.ConfigBase()
        base.set_config_arguments(cfg)
        cfg.add_argment('url', type=str, default='ws://localhost:4455/')
        cfg.add_argment('password', type=str, default=None)
        cfg.add_argment('source_name', type=str, default=None)
        cfg.parse(data)
        return cfg

    def __init__(self, ctx, cfg=None):
        super().__init__(ctx=ctx, cfg=cfg)
        self.logger = logging.getLogger(f'obsws({self.cfg.location})')
        self._pending_text = None
        self._sending_task = None
        self.ws = None
        self.connect_to(cfg.src)

    async def _ws_connect(self):
        self.ws = simpleobsws.WebSocketClient(url=self.cfg.url, password=self.cfg.password)
        await self.ws.connect()
        if not await self.ws.wait_until_identified():
            self.ws.disconnect()
            self.ws = None
            raise ConnectionError('Identification error')

    async def update(self, slide, args):
        if not slide:
            text = ''
        elif isinstance(slide, str):
            text = slide
        else:
            text = str(slide)

        self._send_text(text)

    def _send_text(self, text):
        self._pending_text = text
        if not self._sending_task:
            self._sending_task = asyncio.create_task(self._send_text_routine())

    async def _send_text_routine(self):
        self.logger.debug('Starting text sending routine...')
        sent_text = object()
        while self._pending_text != sent_text:
            text = self._pending_text

            try:
                self.logger.debug('Attempting to send "%s"...', text)
                await self._send_request(simpleobsws.Request('SetInputSettings', {
                    'inputName': self.cfg.source_name,
                    'inputSettings': {'text': text}
                }))
                sent_text = text
            except Exception as e:
                self.logger.warning('Could not send text. %s', e)
                break

        self._sending_task = None
        self.logger.debug('Ending text sending routine. The text was "%s"', text)

    async def _send_request(self, req, retry=2):
        while retry > 0:
            retry -= 1

            if not self.ws:
                try:
                    await self._ws_connect()
                except Exception as e:
                    self.logger.warning('Could not connect to %s. %s', self.cfg.url, e)
                    self.ws = None
            if not self.ws:
                if retry > 0:
                    self.logger.info('Retrying to connect...')
                continue

            try:
                res = await self.ws.call(req)
                if res.ok():
                    return res.responseData
                self.logger.warning('%s: %d: %s', res.requestType,
                                    res.requestStatus.code, res.requestStatus.comment)

            except Exception as e:
                self.logger.warning('Failed to send text. %s', e)

                try:
                    await self.ws.disconnect()
                except Exception as e1:
                    self.logger.warning('Disconnect failed. %s', e1)
                self.ws = None

                if retry == 0:
                    raise e
                self.logger.info('Retrying...')



import sys
import types
import unittest
from unittest.mock import MagicMock, patch, AsyncMock

try:
    import uno
except ImportError:
    mock_uno = types.SimpleNamespace(getComponentContext=lambda: None)
    sys.modules['uno'] = mock_uno

from slidetextbridge.plugins import impress


class TestImpressCapture(unittest.IsolatedAsyncioTestCase):
    def test_type_name(self):
        self.assertEqual(impress.ImpressCapture.type_name(), 'impress')

    async def _test_loop(self, uno_shapes_list):
        ctx = MagicMock()

        uno_ctx = MagicMock()

        resolver = MagicMock()
        uno_ctx.ServiceManager.createInstanceWithContext.return_value = resolver

        uno_inst = MagicMock()
        resolver.resolve.return_value = uno_inst

        desktop = MagicMock()
        uno_inst.ServiceManager.createInstanceWithContext.return_value = desktop

        component = MagicMock()
        desktop.getCurrentComponent.return_value = component

        presentation = MagicMock()
        component.getPresentation.return_value = presentation

        controller = MagicMock()
        presentation.getController.return_value = controller

        def _create_uno_slide(u):
            uno_slide = MagicMock()
            uno_slide.__iter__.return_value = u
            return uno_slide
        controller.getCurrentSlide.side_effect = [_create_uno_slide(u) for u in uno_shapes_list]

        cfg = impress.ImpressCapture.config({})
        obj = impress.ImpressCapture(ctx=ctx, cfg=cfg)

        with (patch('asyncio.sleep') as mock_sleep,
              patch('uno.getComponentContext') as MockGetComponentContext):
            mock_sleep.side_effect = [None] * (len(uno_shapes_list) -1) + [StopAsyncIteration]
            MockGetComponentContext.return_value = uno_ctx

            obj.emit = AsyncMock()

            with self.assertRaises(StopAsyncIteration):
                await obj._loop()

            slides = [x.args[0] for x in obj.emit.call_args_list]

        uno_ctx.ServiceManager.createInstanceWithContext.assert_called_once_with(
                'com.sun.star.bridge.UnoUrlResolver', uno_ctx)
        resolver.resolve.assert_called_once_with(
                'uno:socket,host=localhost,port=2002;urp;StarOffice.ComponentContext')
        uno_inst.ServiceManager.createInstanceWithContext.assert_called_once_with(
                'com.sun.star.frame.Desktop', uno_inst)

        return slides

    async def _test_loop_once(self, uno_shapes):
        slides = await self._test_loop(uno_shapes_list=[uno_shapes])
        return slides[0]

    async def test_texts(self):
        shape1 = MagicMock()
        shape1.Text.getString.return_value = 'text1'
        shape2 = MagicMock()
        shape2.Text.getString.return_value = 'text2'
        uno_shapes = [shape1, shape2]

        slide = await self._test_loop_once(uno_shapes = uno_shapes)
        uno_slide = MagicMock()
        uno_slide.__iter__.return_value = uno_shapes

        self.assertEqual(slide.to_dict(), impress._dict_from_slide(uno_slide))
        self.assertEqual(slide.to_texts(), ['text1', 'text2'])

    async def test_texts_heights(self):
        shape1 = MagicMock()
        shape1.Text.getString.return_value = 'text1'
        shape1.CharHeight = 16
        shape2 = MagicMock()
        shape2.Text.getString.return_value = 'text2'
        shape2.CharHeight = 16
        uno_shapes = [shape1, shape2]

        slide = await self._test_loop_once(uno_shapes = uno_shapes)

        d = slide.to_dict()
        self.assertEqual(len(d['shapes']), 2)
        self.assertEqual(d['shapes'][0]['text'], 'text1')
        self.assertEqual(d['shapes'][0]['char_height'], 16)
        self.assertEqual(d['shapes'][1]['text'], 'text2')
        self.assertEqual(d['shapes'][1]['char_height'], 16)

        # cover to_texts()
        self.assertEqual(str(slide), 'text1\ntext2')

    async def test_2_iterations(self):
        shape1 = MagicMock()
        shape1.Text.getString.return_value = 'text1'
        shape2 = MagicMock()
        shape2.Text.getString.return_value = 'text2'
        uno_shapes = [shape1, shape2]

        slide1, slide2 = await self._test_loop(uno_shapes_list = [
            [shape1, shape2],
            [shape2, shape1]
        ])
        uno_slide = MagicMock()
        uno_slide.__iter__.return_value = uno_shapes

        self.assertEqual(slide1.to_texts(), ['text1', 'text2'])
        self.assertEqual(slide2.to_texts(), ['text2', 'text1'])


if __name__ == "__main__":
    unittest.main()



import sys
import tempfile
import unittest
import logging
from unittest.mock import MagicMock, patch

from slidetextbridge.core.logging import HideSameLogFilter

class _TestFilter(logging.Filter):
    def __init__(self):
        super().__init__()
        self.records = []

    def filter(self, record):
        self.records.append(record)
        return False

class TestHideSameLogFilter(unittest.TestCase):

    def test_all(self):
        logging.basicConfig(level=logging.INFO)
        logger = logging.getLogger(__name__)
        logger.addFilter(HideSameLogFilter(2))

        f = _TestFilter()
        logger.addFilter(f)

        logger.info('A')
        logger.info('B')
        logger.info('B')
        logger.info('C')
        logger.info('D')
        logger.info('B')
        logger.info('D')

        records = [r.msg for r in f.records]
        self.assertEqual(records, ['A', 'B', 'C', 'D', 'B'])

if __name__ == '__main__':
    unittest.main()



'''
Base class definition
'''

import logging
import re

def set_config_arguments(cfg, has_src=True):
    '''
    Set the standard config arguments
    :param cfg:  Instance of ConfigBase
    '''
    cfg.add_argment('type', type=str)
    cfg.add_argment('name', type=str, default=None)
    if has_src:
        cfg.add_argment('src', type=str, default=None)


class PluginBase:
    '''
    The base class for text sources, filters, and sinks.
    '''

    @staticmethod
    def type_name(): # pragma: no cover
        'Return the name of the type'
        raise NotImplementedError()

    @staticmethod
    def config(data): # pragma: no cover
        'Return the config object'
        raise NotImplementedError()

    def __init__(self, ctx, cfg=None):
        self.ctx = ctx
        self.cfg = cfg
        self.name = cfg.name
        self.sinks = []
        if not hasattr(self, 'logger'):
            self.logger = logging.getLogger(f'{__name__}({self.cfg.location})')

    def connect_to(self, name=None, args=None):
        '''
        Connect this instance to the source of the text.
        :param name:  The name of the source
        :param args:  The argument that will be passed to `update` callback
        '''
        try:
            src = self.ctx.get_instance(name=name)
            return src.add_sink(self, args=args)
        except (IndexError, KeyError) as e:
            self.logger.error('Cannot connect to the source. %s', e)
            raise e

    def add_sink(self, sink, args=None):
        '''
        Connect a sink, that will receive the contents of this plugin.
        :param sink:  The instance of PluginBase
        :param args:  Any object
        '''
        self.sinks.append((sink, args))

    async def initialize(self):
        '''
        Callback when the system is started.
        '''

    async def emit(self, slide):
        '''
        Send the new slide to all sinks.
        :param slide:  The new slide.
        '''
        for sink, args in self.sinks:
            try:
                await sink.update(slide, args)
            except Exception as e:
                self.logger.error('Failed to send text to %s. %s', type(sink), e)

    async def update(self, slide, args):
        '''
        The callback function when updating the new slide.
        '''

def _traverse_obj_for_texts(obj: dict, text_paths):
    if isinstance(obj, str):
        yield obj
        return
    if isinstance(obj, (int, float, bool)):
        return
    if isinstance(obj, dict):
        if isinstance(text_paths, dict):
            for key, sub_paths in text_paths.items():
                if key in obj:
                    yield from _traverse_obj_for_texts(obj[key], sub_paths)
        elif isinstance(text_paths, (list, tuple)):
            for key in text_paths:
                if key in obj:
                    yield from _traverse_obj_for_texts(obj[key], text_paths)
        return
    if isinstance(obj, (list, tuple)):
        for item in obj:
            yield from _traverse_obj_for_texts(item, text_paths)

class SlideBase():
    '''
    The base class representing the current slide
    '''

    _default_text_paths = {
            'shapes': {
                'text': None,
            },
    }

    def __init__(self, data=None, parent=None):
        self._dict = data or {}
        if parent or not hasattr(self, 'parent'):
            self.parent = parent

    def set_data(self, data):
        '''
        Update the data for the slide.
        :param data: The new data
        '''
        self._dict = data

    def to_texts(self):
        '''
        List all texts
        :return:  List of strings

        By default, `shapes.[*].text` will be selected.
        To change the search paths, set `text_paths`.
        '''
        if not self._dict:
            return []

        if 'text_paths' in self._dict:
            text_paths = self._dict['text_paths']
        else:
            text_paths = SlideBase._default_text_paths

        try:
            return list(_traverse_obj_for_texts(self._dict, text_paths))
        except (TypeError, KeyError) as e:
            logger = self.parent.logger if self.parent else logging.getLogger('SlideBase')
            logger.error('Failed to convert slide to texts. %s', e)
        return []

    def __str__(self):
        return '\n'.join(self.to_texts())

    @staticmethod
    def convert_object(obj, params=()):
        '''
        A helper method to convert an object to dict type.
        :param params:  List of attributes that will be converted to the dict item.
        Each element of `params` can be a string or a 2-element tuple.
        '''
        d = {}
        def _f(attrname, f_conv=None):
            dest = re.sub('(.)([A-Z])', r'\1_\2', attrname).lower()
            try:
                v = getattr(obj, attrname)
            except: # pylint: disable=W0702
                return
            if f_conv:
                v = f_conv(v)
            d[dest] = v
        for p in params:
            if isinstance(p, tuple):
                _f(p[0], p[1])
            else:
                _f(p)
        return d

    def to_dict(self):
        '''
        Get the dict object so that jmespath can filter it.
        The jmespath filter will then calls the initializer of the class
        with a `data` parameter to create a new slide object.
        :return: dict object representing this object
        '''
        return self._dict or {}



import unittest
import logging
from unittest.mock import MagicMock, patch, AsyncMock
from aiohttp.test_utils import AioHTTPTestCase

from slidetextbridge.plugins import webserver, base

# https://docs.aiohttp.org/en/stable/testing.html#unittest

class TestWebServerEmitter(AioHTTPTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls._orig_log_level = {}
        for name in ('aiohttp', 'asyncio'):
            logger = logging.getLogger(name)
            cls._orig_log_level[name] = logger.level
            logger.setLevel(logging.WARNING)

    @classmethod
    def tearDownClass(cls):
        for name, level in cls._orig_log_level.items():
            logging.getLogger(name).setLevel(level)

        super().tearDownClass()

    def test_type_name(self):
        self.assertEqual(webserver.WebServerEmitter.type_name(), 'webserver')

    def test_config(self):
        cfg_data = {'host': '192.0.2.3', 'port': 8081}
        cfg = webserver.WebServerEmitter.config(cfg_data)
        self.assertEqual(cfg.host, '192.0.2.3')
        self.assertEqual(cfg.port, 8081)

    async def get_application(self):
        self.obj = await self._create_mocked_obj(cfg_data={
            'host': '192.0.2.3',
            'port': 8081,
        })
        return self.obj.app

    @staticmethod
    async def _create_mocked_obj(cfg_data):
        ctx = MagicMock()

        cfg = webserver.WebServerEmitter.config(cfg_data)

        mockAppRunner = MagicMock()
        mockAppRunner.return_value = MagicMock()
        mockAppRunner.return_value.setup = AsyncMock()
        mockTCPSite = MagicMock()
        mockTCPSite.return_value = MagicMock()
        mockTCPSite.return_value.start = AsyncMock()

        with patch.multiple(
                'aiohttp.web',
                AppRunner=mockAppRunner,
                TCPSite=mockTCPSite,
        ):
            obj = webserver.WebServerEmitter(ctx=ctx, cfg=cfg)
            obj.logger = MagicMock()
            await obj.initialize()
            obj.logger.info.assert_called_once_with('Listening on %s:%d', cfg_data['host'], cfg_data['port'])

        mockAppRunner.assert_called_once_with(obj.app)
        mockAppRunner.return_value.setup.assert_called_once_with()
        mockTCPSite.assert_called_once_with(
                mockAppRunner.return_value,
                obj.cfg.host, obj.cfg.port)
        mockTCPSite.return_value.start.assert_called_once_with()

        return obj

    async def test_get(self):
        async with self.client.request('GET', '/') as resp:
            self.assertEqual(resp.status, 200)
            text = await resp.text()
            self.assertIn('<html>', text)

    async def test_ws_text(self):

        # no text are pending
        async with self.client.ws_connect('/ws/text') as ws:
            await self.obj.update('first text', None)
            self.assertEqual(await ws.receive_str(), 'first text')

        # pending text will be sent soon
        await self.obj.update('pending text', None)
        async with self.client.ws_connect('/ws/text') as ws:
            await ws.send_str('ignored text')
            await self.obj.update('updated text', None)
            self.assertEqual(await ws.receive_str(), 'pending text')
            self.assertEqual(await ws.receive_str(), 'updated text')

    async def test_ws_shape_and_clear(self):
        async with self.client.ws_connect('/ws/text') as ws:
            # The SlideBase instance will be converted to a string.
            await self.obj.update(base.SlideBase({'shapes': {'text': 'shape text'}}), None)
            self.assertEqual(await ws.receive_str(), 'shape text')

            # Also tests to clear the text.
            await self.obj.update(None, None)
            self.assertEqual(await ws.receive_str(), '')

class TestWebServerEmitter_CustomContents(TestWebServerEmitter):

    async def get_application(self):
        self.obj = await self._create_mocked_obj(cfg_data={
            'host': '192.0.2.3',
            'port': 8081,
            'index_html': '<!-- index html -->',
            'script_js': '/* script data */',
            'style_css': '/* style data */',
        })
        return self.obj.app

    async def test_get(self):
        async with self.client.request('GET', '/') as resp:
            self.assertEqual(resp.status, 200)
            text = await resp.text()
            self.assertEqual(text, '<!-- index html -->')

        async with self.client.request('GET', '/script.js') as resp:
            self.assertEqual(resp.status, 200)
            text = await resp.text()
            self.assertEqual(text, '/* script data */')

        async with self.client.request('GET', '/style.css') as resp:
            self.assertEqual(resp.status, 200)
            text = await resp.text()
            self.assertEqual(text, '/* style data */')



if __name__ == '__main__':
    unittest.main()



'''
Send text through HTTP
'''

import logging
import os
from aiohttp import web
from slidetextbridge.core import config
from . import base


def _get_data_path(file):
    module_dir = os.path.dirname(os.path.abspath(__file__))
    return module_dir + '/data/webserver/' + file


class _GetHandler:
    def __init__(self, content=None, path=None, content_type=None, logger=None):
        if content:
            self.content = content
        elif path:
            self.content = None
            self.load(path)
        elif logger: # pragma: no cover
            logger.error('_GetHandler: Need path or content')

        self.content_type = content_type or 'text/html'

    def load(self, path):
        'Load the content from the given path'
        with open(path, encoding='utf-8') as fr:
            self.content = fr.read()

    async def handler(self, request):
        'The handler compatible with aiohttp'
        # pylint: disable=unused-argument
        return web.Response(text=self.content, content_type=self.content_type)


class WebServerEmitter(base.PluginBase):
    '''
    HTTP server to send text through HTTP
    '''
    @staticmethod
    def type_name():
        return 'webserver'

    @staticmethod
    def config(data):
        'Return the config object'
        cfg = config.ConfigBase()
        base.set_config_arguments(cfg)
        cfg.add_argment('host', type=str, default='localhost')
        cfg.add_argment('port', type=int, default=8080)
        cfg.add_argment('uri_index', type=str, default='/index.html')
        cfg.add_argment('uri_script', type=str, default='/script.js')
        cfg.add_argment('uri_style', type=str, default='/style.css')
        cfg.add_argment('index_html', type=str, default='')
        cfg.add_argment('script_js', type=str, default='')
        cfg.add_argment('style_css', type=str, default='')
        cfg.add_argment('theme', type=str, default='default')
        cfg.parse(data)
        return cfg

    def __init__(self, ctx, cfg=None):
        super().__init__(ctx=ctx, cfg=cfg)
        self.logger = logging.getLogger(f'webserver({self.cfg.location})')
        self.app = None
        self.clients = set()
        self._last_text = None
        self.connect_to(cfg.src)

    def _fmt(self, text):
        if not text:
            return None
        return text \
                .replace('{uri_script}', self.cfg.uri_script) \
                .replace('{uri_style}', self.cfg.uri_style)

    async def _handle_ws(self, request):
        # pylint: disable=unused-argument
        self.logger.info('Starting %s remote=%s', request.rel_url, request.remote)
        try:
            ws = web.WebSocketResponse()
            await ws.prepare(request)
            self.clients.add(ws)

            if self._last_text:
                await self._send_text(ws, self._last_text)

            async for _ in ws:
                pass
        except Exception as e:
            self.logger.error('%s: Failed to process ws message. %s', request.remote, e)
        finally:
            self.clients.discard(ws)

        self.logger.info('Closing %s remote=%s', request.rel_url, request.remote)
        return ws

    async def initialize(self):
        self.app = web.Application()
        theme_dir = _get_data_path(self.cfg.theme) if self.cfg.theme else '.'
        index_handler = _GetHandler(
                content=self._fmt(self.cfg.index_html),
                path=theme_dir + '/index.html',
                logger=self.logger
        )
        script_handler = _GetHandler(
                content=self._fmt(self.cfg.script_js),
                path=theme_dir + '/script.js',
                content_type='text/javascript',
                logger=self.logger
        )
        style_handler = _GetHandler(
                content=self._fmt(self.cfg.style_css),
                path=theme_dir + '/style.css',
                content_type='text/css',
                logger=self.logger
        )
        self.app.add_routes([
            web.get('/', index_handler.handler),
            web.get(self.cfg.uri_index, index_handler.handler),
            web.get(self.cfg.uri_script, script_handler.handler),
            web.get(self.cfg.uri_style, style_handler.handler),
            web.get('/ws/text', self._handle_ws),
        ])
        runner = web.AppRunner(self.app)
        await runner.setup()
        site = web.TCPSite(runner, self.cfg.host, self.cfg.port)
        await site.start()
        self.logger.info('Listening on %s:%d', self.cfg.host, self.cfg.port)

    async def update(self, slide, args):
        if not slide:
            text = ''
        elif isinstance(slide, str):
            text = slide
        else:
            text = str(slide)

        self._last_text = text
        for ws in self.clients:
            await self._send_text(ws, text)

    async def _send_text(self, ws, text):
        try:
            await ws.send_str(text)
        except Exception as e:
            self.logger.warning('Failed to send text to ws. %s', e)



'''
Filter with JMESPath
'''

import jmespath
from slidetextbridge.core import config
from . import base

def _convert_text_paths(orig_obj):
    if 'text_paths' not in orig_obj:
        return ['shapes', 'text']

    def _traverse_dict(d):
        if isinstance(d, dict):
            for key, value in d.items():
                yield key
                if value:
                    yield from _traverse_dict(value)
        if isinstance(d, (list, tuple)):
            yield from d

    return list(_traverse_dict(orig_obj['text_paths']))


class JMESPathFilter(base.PluginBase):
    '''
    Filter shapes with JMESPath
    '''
    @staticmethod
    def type_name():
        return 'jmespath'

    @staticmethod
    def config(data):
        'Return the config object'
        cfg = config.ConfigBase()
        base.set_config_arguments(cfg)
        cfg.add_argment('filter', type=str)
        cfg.parse(data)
        return cfg

    def __init__(self, ctx, cfg=None):
        super().__init__(ctx=ctx, cfg=cfg)
        self.connect_to(cfg.src)
        self.jmespath_filter = jmespath.compile(cfg.filter)

    async def update(self, slide, args):
        orig_obj = slide.to_dict()
        new_obj = self.jmespath_filter.search(orig_obj)

        if not isinstance(new_obj, dict):
            new_obj = {'shapes': new_obj}

        new_obj['text_paths'] = _convert_text_paths(orig_obj)

        for key, value in orig_obj.items():
            if key in new_obj:
                continue
            new_obj[key] = value

        slide = base.SlideBase(data=new_obj)
        await self.emit(slide)
